#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "hash.h"

typedef struct _HashNode
{
    void *user;
    struct _HashNode *next;
} HashNode;


typedef struct _HashHead
{
    unsigned long size;
    HashNode **nodes;
    HashFunc hash_func;
    HashCmp cmp_func;
} HashHead;

typedef HashNode *pHashNode;

/* inline void inserthere(unsigned long,void *); */

void *HashInit(HashFunc func1,HashCmp func2,unsigned long size)
{
    HashHead *p;

    p= (HashHead *) malloc(sizeof(HashHead));
    p->nodes= calloc(size,sizeof(pHashNode));
    if (!p->nodes)
    {
	free(p);
	return 0;
    }
    memset(p->nodes,0,size*sizeof(HashNode*));
    p->size= size;
    p->hash_func= func1;
    p->cmp_func= func2;
    return p;
}


inline void inserthere(HashHead *head,unsigned long i,void *user)
{
    HashNode *p;

    p= (HashNode *) malloc(sizeof(HashNode));
    p->user= user;
    p->next= head->nodes[i];
    head->nodes[i]= p;
}


void *HashInsert(void *x,void *user)
{
    HashHead *head= (HashHead *) x;
    HashNode *p;
    unsigned long i;
    unsigned long h;

    h= (*(head->hash_func))(user);
    i= h % head->size;
    p= head->nodes[i];
    if (!p) 
    {
	inserthere(head,i,user);
	return 0;
    }
    while (p)
    {
	if ((*(head->cmp_func))(user,p->user)) return p->user;
	p= p->next;
    }
    inserthere(head,i,user);
    return 0;
}


void *HashDelete(void *x,void *user)
{
    HashHead *head= (HashHead *) x;
    HashNode *p,*q;
    unsigned long i;

    i= (*(head->hash_func))(user) % head->size;
    p= head->nodes[i];
    q= 0;
    while (p)
    {
	if ((*(head->cmp_func))(user,p->user)) 
	{
	    void *temp= p->user;
	    if (q) q->next= p->next;
	    else head->nodes[i]= p->next;
	    temp= p->user;
	    free(p);
	    return temp;
	}
	q= p; p= p->next;
    }
    return 0;
}


void *HashSearch(void *x,void *user)
{
    HashHead *head= (HashHead *) x;
    HashNode *p;
    unsigned long i;

    i= (*(head->hash_func))(user) % head->size;
    p= head->nodes[i];
    while (p)
    {
	if ((*(head->cmp_func))(user,p->user)) return p->user;
	p= p->next;
    }
    return 0;
}


void HashDestroy(void *x,HashDest func)
{
    HashHead *head= (HashHead *) x;
    HashNode *p,*q;
    unsigned long i;

    for (i=0;i<head->size;i++)
    {
	p= head->nodes[i];
	while (p)
	{
	    (*func)(p->user);
	    q= p->next;
	    free(p);
	    p= q;
	}
    }
    free(head->nodes);
    free(head);
}


int HashTraverse(void *x,HashTrav func)
{
    HashHead *head= (HashHead *) x;
    HashNode *p;
    unsigned long i;

    for (i=0;i<head->size;i++)
    {
	p= head->nodes[i];
	while (p)
	{
	    if (!(*func)(p->user)) return 0;
	    p= p->next;
	}
    }
    return 1;
}


void HashPrintCollisions(void *x)
{
    HashHead *head= (HashHead *) x;
    HashNode *p;
    unsigned long i,j,total,most;

    total= 0;
    most= 0;
    for (i=0;i<head->size;i++)
    {
	p= head->nodes[i];
	for (j=0;p;j++,p=p->next) ;
	if (j>1) 
	{
	    printf("Box %d: %d collisions\n",i,j);
	    total+= j;
	    if (j>most) most= j;
	}
    }
    printf("Total Boxes:      %d\n",head->size);
    printf("Total Collisions: %d\n",total);
    printf("Most Collisions:  %d\n",most);
}
	
    
