#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wff.h"
#include "hash.h"


typedef struct _Numbered
{
    void *p;
    int n;
} Numbered;


static Numbered *newNumbered(void *, int);
static void deleteNumbered(Numbered *);
static int numberedHash1(Numbered *);
static int numberedCmp1(Numbered *, Numbered *);
static int numberedHash2(Numbered *);
static int numberedCmp2(Numbered *, Numbered *);

static WfTriple *getTriple(void *, int, int);
static int getIndex(void *, WfTriple *);
static int maybeInsert(void *, WfTriple *, int);

static WfTriple **increaseArray(WfTriple **, int);

static char line[1024];
static char line2[1024];

WfObject *WfReadFile(char *name)
{
    FILE *fp = fopen(name,"r");
    void *vHash;
    void *vnHash;
    void *vtHash;
    void *fHash;
    char *delim = " \n\t";
    WfObject *obj;
    WfTriple *t;
    Numbered *h;
    WfFace *f;
    int i;

    if (!fp)
	return 0;
    obj = WfNewObject(name);
    vHash = HashInit((HashFunc)numberedHash1, (HashCmp) numberedCmp1, 3000);
    vnHash = HashInit((HashFunc)numberedHash1, (HashCmp) numberedCmp1, 3000);
    vtHash = HashInit((HashFunc)numberedHash1, (HashCmp) numberedCmp1, 3000);
    fHash = HashInit((HashFunc)numberedHash1, (HashCmp) numberedCmp1, 3000);
    fgets(line, 1023, fp);
    while (!feof(fp))
    {
	char *tok;

	tok = strtok(line, delim);

	if (!tok) ;
	else if (!strcmp(tok, "v"))
	{
	    static int first = 1;
	    double s[3];
	    s[0] = atof( strtok(0,delim) );
	    s[1] = atof( strtok(0,delim) );
	    s[2] = atof( strtok(0,delim) );
	    t = WfNewTriplev( s );
	    if( first )
	    {
		printf( "%f %f %f\n", t->tr[0], t->tr[1], t->tr[2] );
		first = 0;
	    }
	    h = newNumbered(t, ++(obj->numVerts));
	    HashInsert(vHash, h);
	}
	else if (!strcmp(tok, "vn"))
	{
	    double s[3];
	    s[0] = atof( strtok(0,delim) );
	    s[1] = atof( strtok(0,delim) );
	    s[2] = atof( strtok(0,delim) );
	    t = WfNewTriplev( s );
	    h = newNumbered(t, ++(obj->numNormals));
	    HashInsert(vnHash, h);
	}
	else if (!strcmp(tok, "vt"))
	{
	    double s[3];
	    s[0] = atof( strtok(0,delim) );
	    s[1] = atof( strtok(0,delim) );
	    s[2] = atof( strtok(0,delim) );
	    t = WfNewTriplev( s );
	    h = newNumbered(t, ++(obj->numTextures));
	    HashInsert(vtHash, h);
	}
	else if (!strcmp(tok, "f"))
	{
	    i = 0;
	    f = WfNewFace();
	    h = newNumbered(f, ++(obj->numFaces));
	    HashInsert(fHash, h);

	    /* go through it once to determine how many there are - yuck! */
	    tok[1] = ' ';
	    strcpy(line2, line);
	    strtok(line2, delim);
	    while ((tok = strtok(0, delim)))
		f->numVerts++;

	    /* now parse them */
	    f->verts = (WfTriple **) calloc(f->numVerts, sizeof(WfTriple *));
	    f->normals = (WfTriple **) calloc(f->numVerts, sizeof(WfTriple *));
	    f->textures = (WfTriple **) calloc(f->numVerts, sizeof(WfTriple *));
	    strtok(line, delim); /* skip first one */
	    while ((tok = strtok(0, delim)))
	    {
		int d = (int) strtol(tok, &tok, 10);

		f->verts[i] = getTriple(vHash, d, obj->numVerts);

		/* get texture */
		if (tok[0]=='/' && tok[1]!='/')
		{
		    d = (int) strtol(tok+1, &tok, 10);
		    f->textures[i] = getTriple(vtHash, d, obj->numTextures);
		    f->flags |= WF_WITH_TEXTURES;
		}
		else if (tok[0]=='/')
		    tok++;

		/* get normal */
		if (tok[0]=='/' && tok[1]!='\0')
		{
		    d = (int) strtol(tok+1, 0, 10);
		    f->normals[i] = getTriple(vnHash, d, obj->numNormals);
		}
		i++;
	    }
	}
	/* else discard */
	fgets(line, 1023, fp);
    }
    fclose(fp);

    /* make arrays */
    obj->verts = (WfTriple **) calloc(obj->numVerts, sizeof(WfTriple *));
    obj->normals = (WfTriple **) calloc(obj->numNormals, sizeof(WfTriple *));
    obj->textures = (WfTriple **) calloc(obj->numTextures, sizeof(WfTriple *));
    obj->faces = (WfFace **) calloc(obj->numFaces, sizeof(WfFace *));

    /* fill arrays */
    for (i=1; i<=obj->numVerts; i++)
	obj->verts[i-1] = getTriple(vHash, i, 0);
    for (i=1; i<=obj->numNormals; i++)
	obj->normals[i-1] = getTriple(vnHash, i, 0);
    for (i=1; i<=obj->numTextures; i++)
	obj->textures[i-1] = getTriple(vtHash, i, 0);
    for (i=1; i<=obj->numFaces; i++)
	obj->faces[i-1] = (WfFace *) getTriple(fHash, i, 0);

    /* delete hashes */
    HashDelete(vHash, (HashDest) deleteNumbered);
    HashDelete(vnHash, (HashDest) deleteNumbered);
    HashDelete(vtHash, (HashDest) deleteNumbered);
    HashDelete(fHash, (HashDest) deleteNumbered);
    return obj;
}


int WfWriteFile(char *name, WfObject *obj)
{
    FILE *f = name ? fopen(name, "w") : stdout;
    void *vHash, *nHash, *tHash;
    int i;

    if (!f)
	return 0;

    vHash = HashInit((HashFunc)numberedHash2, (HashCmp) numberedCmp2, 3000);
    nHash = HashInit((HashFunc)numberedHash2, (HashCmp) numberedCmp2, 3000);
    tHash = HashInit((HashFunc)numberedHash2, (HashCmp) numberedCmp2, 3000);
    if (obj->fileName)
	fprintf(f, "# %s - WaveFront Object file generated by WFFLib\n#\n",
	    obj->fileName);
    else
	fprintf(f, "# WaveFront Object file generated by WFFLib\n#\n");

    fprintf(f, "# Vertices\n");
    for (i=0; i<obj->numVerts; i++)
    {
	double *v = obj->verts[i]->tr;
	fprintf(f, "v %g %g %g\n", v[0], v[1], v[2]);
	HashInsert(vHash, newNumbered(obj->verts[i], i+1));
    }

    if (obj->numNormals)
    {
	fprintf(f, "#\n# Normals\n");
	for (i=0; i<obj->numNormals; i++)
	{
	    double *v = obj->normals[i]->tr;
	    fprintf(f, "vn %g %g %g\n", v[0], v[1], v[2]);
	    HashInsert(nHash, newNumbered(obj->normals[i], i+1));
	}
    }

    if (obj->numTextures)
    {
	fprintf(f, "#\n# Textures\n");
	for (i=0; i<obj->numTextures; i++)
	{
	    double *v = obj->textures[i]->tr;
	    fprintf(f, "vt %g %g %g\n", v[0], v[1], v[2]);
	    HashInsert(tHash, newNumbered(obj->textures[i], i+1));
	}
    }

    if (obj->numFaces)
    {
	fprintf(f, "#\n# Faces\n");
	for (i=0; i<obj->numFaces; i++)
	{
	    WfFace *p = obj->faces[i];
	    int j;

	    if (!p->numVerts)
		continue;
	    fputs("f ", f);
	    for (j=0; j<p->numVerts; j++)
	    {
		fprintf(f, "%d/", getIndex(vHash, p->verts[j]));
		if (p->textures && p->textures[j])
		    fprintf(f, "%d/", getIndex(tHash, p->textures[j]));
		else
		    fputc((int)'/', f);
		if (p->normals && p->normals[j])
		    fprintf(f, "%d ", getIndex(nHash, p->normals[j]));
		else
		    fputc((int)' ', f);
	    }
	    fputc((int)'\n', f);
	}
    }

    HashDelete(vHash, (HashDest) deleteNumbered);
    HashDelete(nHash, (HashDest) deleteNumbered);
    HashDelete(tHash, (HashDest) deleteNumbered);
    if (name)
	fclose(f);
    return 1;
}


WfObject *WfNewObject(char *name)
{
    WfObject *p = (WfObject *) malloc(sizeof(WfObject));
    p->fileName = name ? strdup(name) : 0;
    p->numVerts = 0;
    p->numNormals = 0;
    p->numTextures = 0;
    p->numFaces = 0;
    p->verts = 0;
    p->normals = 0;
    p->textures = 0;
    p->faces = 0;
    return p;
}


void WfDeleteObject(WfObject *p)
{
    int i;

    if (p->fileName)
	free(p->fileName);

    /* free subcomponents */
    for (i=0; i<p->numVerts; i++)
	WfDeleteTriple(p->verts[i]);
    for (i=0; i<p->numNormals; i++)
	WfDeleteTriple(p->normals[i]);
    for (i=0; i<p->numTextures; i++)
	WfDeleteTriple(p->textures[i]);
    for (i=0; i<p->numFaces; i++)
	WfDeleteFace(p->faces[i]);

    /* free arrays */
    if (p->verts)
	free(p->verts);
    if (p->normals)
	free(p->normals);
    if (p->textures)
	free(p->textures);
    if (p->faces)
	free(p->faces);

    free(p);
}


WfTriple *WfNewTriple(double x,double y,double z)
{
    WfTriple *p = (WfTriple *) malloc(sizeof(WfTriple));

    p->tr[0] = x; p->tr[1] = y; p->tr[2] = z;
    return p;
}


WfTriple *WfNewTriplev(double *a)
{
    return WfNewTriple(a[0], a[1], a[2]);
}


void WfDeleteTriple(WfTriple *p)
{ 
    free(p); 
}


WfFace *WfNewFace()
{
    WfFace *p = (WfFace *) malloc(sizeof(WfFace));

    p->flags = 0;
    p->numVerts = 0;
    p->verts = (WfTriple **) calloc(3, sizeof(WfTriple *));
    p->normals = (WfTriple **) calloc(3, sizeof(WfTriple *));
    p->textures = (WfTriple **) calloc(3, sizeof(WfTriple *));
    return p;
}

void WfDeleteFace(WfFace *f)
{
    if (f->verts)
	free(f->verts);
    if (f->normals)
	free(f->normals);
    if (f->textures)
	free(f->textures);
    free(f);
}


void WfGetCenter(double c[3], WfFace *f)
{
    int i;
    double inv = 1./f->numVerts;

    c[0] = 0; c[1] = 0; c[2] = 0;
    for (i=0; i<f->numVerts; i++)
    {
	c[0] += f->verts[i]->tr[0];
	c[1] += f->verts[i]->tr[1];
	c[2] += f->verts[i]->tr[2];
    }
    c[0] *= inv;
    c[1] *= inv;
    c[2] *= inv;
}


void WfAddVertex(WfFace *f, WfTriple *v, WfTriple *n, WfTriple *t)
{
    int i = f->numVerts++;

    if (i >= 3)
    {
	f->verts = increaseArray(f->verts, i);
	f->normals = increaseArray(f->normals, i);
	f->textures = increaseArray(f->textures, i);
    }
    f->verts[i] = v;
    f->normals[i] = n;
    f->textures[i] = t;
    if (n)
	f->flags |= WF_WITH_NORMALS;
    if (t)
	f->flags |= WF_WITH_TEXTURES;
}


Numbered *newNumbered(void *q, int n)
{
    Numbered *p = (Numbered *) malloc(sizeof(Numbered));

    p->p = q;
    p->n = n;
    return p;
}


void deleteNumbered(Numbered *p)
{
    free(p);
}


int numberedHash1(Numbered *p)
{
    return p->n;
}


int numberedCmp1(Numbered *p, Numbered *q)
{
    return p->n == q->n;
}


int numberedHash2(Numbered *p)
{
    int i = (int) p->p;
    if (i<0) i *= -1;
    return i;
}

int numberedCmp2(Numbered *p, Numbered *q)
{
    return p->p == q->p;
}


static WfTriple *getTriple(void *hash, int d, int total)
{
    Numbered numbered;
    Numbered *h;

    if (d > 0)
	numbered.n = d;
    else
	numbered.n = total + d + 1;
    h = (Numbered *) HashSearch(hash, &numbered);
    return (WfTriple *) h->p;
}


static int getIndex(void *hash, WfTriple *p)
{
    Numbered numbered;
    Numbered *h;

    numbered.p = p;
    h = (Numbered *) HashSearch(hash, &numbered);
    return h->n;
}


static int maybeInsert(void *hash, WfTriple *p, int i)
{
    Numbered numbered;
    Numbered *h;

    numbered.p = p;
    if (!HashSearch(hash, &numbered))
    {
	h = newNumbered(p, ++i);
	HashInsert(hash, p);
    }
    return i;
}


static WfTriple **increaseArray(WfTriple **a, int size)
{
    WfTriple **a2 = (WfTriple **) calloc(size+1, sizeof(WfTriple *));
    memcpy(a2, a, size*sizeof(WfTriple *));
    free(a);
    return a2;
}


