#include <stdio.h>
#include <math.h>
#include <wff.h>


static void addNormals( WfObject * );
static WfTriple *getNormal( WfFace * );

int main( int argc, char **argv )
{
    WfObject *obj;

    if( argc != 2 )
    {
	fprintf( stderr, "Usage: nobj file\n" );
	fprintf( stderr, "Reads a WaveFront object file, and prints out\n" );
	fprintf( stderr, "on stdout the same object with normals added.\n" );
	exit( 1 );
    }

    obj = WfReadFile( argv[1] );
    if( !obj )
    {
	fprintf( stderr, "%s not a WaveFront Object file\n", argv[1] );
	exit( 1 );
    }
    if( obj->numNormals )
    {
	fprintf( stderr, "File already has normals\n" );
	exit( 1 );
    }
    addNormals( obj );
    WfWriteFile( 0, obj );
    return 0;
}


/* add normals to the object in-place */
void addNormals( WfObject *obj )
{
    int i;
    WfTriple **normals = (WfTriple **) calloc( obj->numFaces,
	sizeof(WfTriple *) );

    obj->numNormals = obj->numFaces;
    if( obj->normals )
	free( obj->normals );
    obj->normals = normals;
    for( i=0; i<obj->numFaces; i++ )
    {
	WfFace *f = obj->faces[i];
	int j;

	normals[i] = getNormal( f );
	if( f->normals )
	    free( f->normals );
	f->normals = (WfTriple **) calloc( f->numVerts, sizeof(WfTriple *) );
	for( j=0; j<f->numVerts; j++ )
	    f->normals[j] = normals[i];
    }

}


static WfTriple *getNormal( WfFace *f )
{
    WfTriple *normal = WfNewTriple( 0, 0, 0 );
    double *n = normal->tr;
    double mag;
    int i;

    for( i=0; i<f->numVerts; i++ )
    {
	int j = (i+1) % f->numVerts;
	double *vi = f->verts[i]->tr;
	double *vj = f->verts[j]->tr;
	n[0] += ( vi[1] - vj[1] )*( vi[2] + vj[2] );
	n[1] += ( vi[2] - vj[2] )*( vi[0] + vj[0] );
	n[2] += ( vi[0] - vj[0] )*( vi[1] + vj[1] );
    }
    mag = sqrt( n[0]*n[0] + n[1]*n[1] + n[2]*n[2] );
    n[0] /= mag;
    n[1] /= mag;
    n[2] /= mag;
    return normal;
}
